(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith Omega Wellfounded.

Set Implicit Arguments.

(* Dependent recursion principle for Acc, same as Acc_inv_dep but for sort Type
   instead of sort Prop *)

Scheme Acc_drect := Induction for Acc Sort Type.

Section Acc_eq.

  (* Proof that Fix_F is extensionnal *)

  Variable (X : Type) (R : X -> X -> Prop).
  
  Fact Acc_irrefl x : R x x -> ~ Acc R x.
  Proof.
    intros H1 H2; revert H2 H1.
    induction 1 as [ x _ Hx ]; intros; apply (Hx x); auto.
  Qed.

  (* This is a weaker notion than @eq (Acc R x) but sufficient
     to establish extensinality properties *)

  Inductive Acc_eq : forall x, Acc R x -> Acc R x -> Prop := 
    | in_Acc_eq : forall x H1 H2, (forall y Hy, Acc_eq (H1 y Hy) (H2 y Hy)) -> Acc_eq (Acc_intro x H1) (Acc_intro _ H2).

  (* unless dependent functional extensionality is assumed, it is
     not possible to show that there is a unique accessibility proof
     BUT it is easy to show that they are unique upto Acc_eq ... *)

  Fact Acc_eq_total x (H1 H2 : Acc R x) : Acc_eq H1 H2.
  Proof.
    revert x H1 H2.
    induction H1 as [ x H1 IH ] using Acc_inv_dep; intros [ H2 ].
    constructor; intros; apply IH.
  Qed.

  Variable P : X -> Type.
  Variable F : forall x, (forall y, R y x -> P y) -> P x.

  Variable (E : forall x, P x -> P x -> Prop).

  Hypothesis
    F_ext :
      forall x (f g:forall y, R y x -> P y),
        (forall y (p:R y x), E (f y p) (g y p)) -> E (F f) (F g).

  (* This one generalizes Fix_F_inv from Wf in that it does not assume wellfoundedness *)

  Lemma Fix_F_ext : forall x (r s : Acc R x), E (Fix_F _ F r) (Fix_F _ F s).
  Proof.
    intros ? r s; generalize (Acc_eq_total r s).
    induction 1 as [ ? ? ? ? IH ]; apply F_ext, IH.
  Qed.

  (* This one generalizes Fix_F_eq for arbitrary relations *) 

  Theorem Fix_F_equiv x (r : Acc R x) : E (Fix_F _ F r) 
                                          (F (fun y Hy => Fix_F _ F (Acc_inv r _ Hy))).
  Proof.
    rewrite Fix_F_eq; apply Fix_F_ext.
  Qed.

  Hypothesis Rwf : well_founded R.

  (* This one generalizes Fix_eq for arbitrary relations *)

  Theorem Fix_equiv x : E (Fix Rwf P F x) (F (fun y _ => Fix Rwf P F y)).
  Proof.
    unfold Fix.
    destruct (Rwf x); simpl.
    apply F_ext. 
    intros; apply Fix_F_ext.
  Qed.

End Acc_eq.

Section seq_acc.

  Variables (R : nat -> nat -> Prop).
  Variables (G D K : nat -> Prop) 
            (HK : forall n, K n -> Acc R n)
            (HGD1 : forall n, K n -> G n \/ D n)
            (HGD2 : forall n, G n -> D n -> False)
            (Hrec : forall n, D n -> R (S n) n)
            (HK' : forall x y, K x -> R y x -> K y).

  Lemma seq_acc u : K u -> exists n, u <= n /\ G n /\ forall i, u <= i < n -> D i.
  Proof.
    intros H.
    generalize (HK H); intros H'; revert H' H.
    induction 1 as [ u Hu IH ]; intros Hk.
    destruct (@HGD1 _ Hk) as [ H | H ].
    exists u.
    repeat split; auto.
    intros ? ?; omega.
    destruct (IH _ (Hrec H)) as (n & ? & ? & ?).
    apply HK' with (1 := Hk).
    apply Hrec; auto.
    exists n.
    repeat split; auto.
    omega.
    intros i H3.
    destruct (eq_nat_dec u i).
    subst; auto.
    apply H2.
    omega.
  Qed.

End seq_acc.

