(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Omega Max Relations Wellfounded.

Require Import utils tree.

Set Implicit Arguments.

(* Dependent recursion principle for Acc, same as Acc_inv_dep but for sort Type
   instead of sort Prop 

Scheme Acc_drect := Induction for Acc Sort Type.

Section Acc_eq.

  (* Proof that Fix_F is extensionnal *)

  Variable (X : Type) (R : X -> X -> Prop).

  Inductive Acc_eq : forall x, Acc R x -> Acc R x -> Prop := 
    | in_Acc_eq : forall x H1 H2, (forall y Hy, Acc_eq (H1 y Hy) (H2 y Hy)) -> Acc_eq (Acc_intro x H1) (Acc_intro _ H2).

  Fact Acc_eq_total x (H1 H2 : Acc R x) : Acc_eq H1 H2.
  Proof.
    revert x H1 H2.
    induction H1 as [ x H1 IH ] using Acc_inv_dep; intros [ H2 ].
    constructor; intros; apply IH.
  Qed.

  Variable P : X -> Type.
  Variable F : forall x, (forall y, R y x -> P y) -> P x.

  Hypothesis
    F_ext :
      forall x (f g:forall y, R y x -> P y),
        (forall y (p:R y x), f y p = g y p) -> F f = F g.

  (* This one generalizes Fix_F_inv in that it does not assume wellfoundedness *)

  Lemma Fix_F_ext : forall x (r s:Acc R x), Fix_F _ F r = Fix_F _ F s.
  Proof.
    intros ? r s; generalize (Acc_eq_total r s).
    induction 1 as [ ? ? ? ? IH ]; apply F_ext, IH.
  Qed.

End Acc_eq.

*)

(* p iterate of a binary relation *)

Section rel_iter.

  Variable (X : Type).
  Implicit Type (R S : X -> X -> Prop).

  Fixpoint rel_iter R p (u v : X) :=
    match p with 
    | 0   => u = v
    | S p => exists k, R u k /\ rel_iter R p k v
  end.

  Fact rel_iter_mono R S p : 
      (forall x y, R x y -> S x y)
    -> forall u v, rel_iter R p u v -> rel_iter S p u v.
  Proof.
    intros H.
    induction p; simpl; intros u v; auto.
    intros (k & ? & ?); exists k; split; auto.
  Qed.

  Fact rel_iter_plus R p q u v : rel_iter R (p+q) u v <-> exists k, rel_iter R p u k /\ rel_iter R q k v.
  Proof.
    revert u v; induction p as [ | p IHp ]; intros u v; simpl.

    split.
    exists u; auto.
    intros (? & [] & ?); auto.

    split.

    intros (k & H1 & H2).
    apply IHp in H2.
    destruct H2 as (j & H3 & H4).
    exists j; split; auto.
    exists k; auto.

    intros (k & (j & H1 & H3) & H2).
    exists j; split; auto.
    apply IHp.
    exists k; auto.
  Qed.
 
  Fact crt_rel_iter R u v : clos_refl_trans _ R u v -> exists i, rel_iter R i u v.
  Proof.
    induction 1 as [ u v | | u v w ? (i & ?) ? (j & ?) ].
    exists 1, v; simpl; auto.
    exists 0; simpl; auto.
    exists (i+j).
    apply rel_iter_plus.
    exists v; auto.
  Qed.

End rel_iter.

Local Reserved Notation "u '-[' p ']->' v" (at level 70).

Section finitary_height.

  Variable (X : Type) (f : X -> list X).

  Notation "u -[ p ]-> v" := (rel_iter (fun a b => In b (f a)) p u v) (at level 70).
  
  Definition fin_ht : forall x, Acc (fun a b => In a (f b)) x -> nat. 
  Proof.
    apply Fix_F with (P := fun _ => nat).
    intros x Hx.
    exact (lmax (map S (list_In_map _ Hx))).
  Defined.

  (* The height does not depend on the proof of accessibility *)

  Fact fin_ht_eq x a1 a2 : @fin_ht x a1 = @fin_ht x a2.
  Proof.
    apply Fix_F_ext with (P := fun _ => nat) (E := fun _ => eq).
    intros; do 2 f_equal.
    apply list_In_map_ext; auto.
  Qed.

  Theorem fin_ht_bound p u v Hu : u -[p]-> v -> p <= @fin_ht u Hu.
  Proof.
    revert p v.
    induction Hu as [ u Hu ] using Acc_inv_dep.
    intros [ | p ] v; simpl.
    intros _; apply le_0_n.
    intros (k & H1 & H2).
    rewrite map_list_In_map.
    apply le_trans with (1 := le_n_S _ _ (H k H1 _ _ H2)).
    apply lmax_prop.
    apply In_list_In_map with (f := fun x Hx => S (fin_ht (Hu x Hx))).
  Qed.

  (* We need the dependent induction principle Acc_drect here 
     The way to compute v is, starting from u, to find w such
     that u -> w and height w is maximal, and start again 
     with w.
   *)

  Theorem fin_ht_lub u Hu : { v | u -[ @fin_ht u Hu ]-> v }.
  Proof.
    induction Hu as [ u Hu IH ] using Acc_drect.
    set (h y Hy := proj1_sig (IH y Hy)).
    assert (Hh : forall y Hy, y -[ fin_ht (Hu y Hy) ]-> h y Hy).
      intros y Hy; apply (proj2_sig (IH y Hy)).
    generalize h Hh; clear IH h Hh; intros h Hh.
    simpl; rewrite map_list_In_map.
    destruct (lmax_inv ((list_In_map (f u) (fun x Hx => S (fin_ht (Hu x Hx))))))
      as [ E | H ].
    exists u; rewrite E; simpl; auto.
    apply list_In_map_In_dec in H.
    2: apply eq_nat_dec.
    destruct H as (x & Hx & H).
    exists (h _ Hx).
    rewrite H.
    exists x; split; auto.
  Qed.

  Theorem fin_ht_dec u Hu v Hv : In u (f v) -> @fin_ht u Hu < @fin_ht v Hv.
  Proof.
    intros H.
    destruct (fin_ht_lub Hu) as (w & Hw).
    apply fin_ht_bound with w.
    exists u; auto.
  Qed.

End finitary_height.

Section finitary_rel.

  Variable (X : Type) (R : X -> X -> Prop) (f : X -> list X) 
           (Hf : forall x y, R x y <-> In x (f y)).

  Let acc_eq x : Acc R x <-> Acc (fun x y => In x (f y)) x.
  Proof.
    split; apply Acc_incl; intros ? ?; apply Hf. 
  Qed.

  Definition fin_rel_ht x (Hx : Acc R x) : nat.
  Proof.
    apply fin_ht with (f := f) (x := x), acc_eq, Hx.
  Defined.

  Fact fin_rel_ht_eq x a1 a2 : @fin_rel_ht x a1 = @fin_rel_ht x a2.
  Proof. apply fin_ht_eq. Qed.

  Notation "u -[ p ]-> v" := (rel_iter (fun x y => R y x) p u v) (at level 70).

  Fact fin_rel_ht_bound p u v Hu : u -[p]-> v -> p <= @fin_rel_ht u Hu.
  Proof.
    intros H.
    apply fin_ht_bound with v.
    revert H; apply rel_iter_mono.
    intros ? ?; apply Hf.
  Qed.

  Fact fin_rel_ht_lub u Hu : { v | u -[ @fin_rel_ht u Hu ]-> v }.
  Proof.   
    destruct (fin_ht_lub _ (proj1 (acc_eq _) Hu))
      as (v & Hv).
    exists v.
    revert Hv.
    unfold fin_rel_ht.
    match goal with |- rel_iter _ ?a _ _ 
                    -> rel_iter _ ?b _ _ => cutrewrite (a=b) end.
    apply rel_iter_mono; intros ? ?; apply Hf.
    apply fin_ht_eq.
  Qed.

  Fact fin_rel_dec u Hu v Hv : R u v -> @fin_rel_ht u Hu < @fin_rel_ht v Hv.
  Proof.
    rewrite Hf; apply fin_ht_dec.
  Qed.

End finitary_rel.

Section finitary_tree.

  Variable (X : Type) (f : X -> list X).

  (* We build the tree of derivation along the relation
     (fun a b => In a (f b)) for accessible elements,
     by induction on the accessibility proof 
   *)

  Definition fin_tree : forall x, Acc (fun a b => In a (f b)) x -> tree X. 
  Proof.
    apply Fix_F.
    exact (fun x Hx => in_tree x (@list_In_map _ (tree X) _ Hx)).
  Defined.

  (* The accessibilty proof must exist but does not matter *)

  Fact fin_tree_eq x H1 H2 : @fin_tree x H1 = @fin_tree x H2.
  Proof.
    apply Fix_F_ext with (P := fun _ => tree X) (E := fun _ => eq).
    intros; f_equal; apply list_In_map_ext; auto.
  Qed.

  Fact fin_tree_fix x ax : 
    @fin_tree x ax = in_tree x (list_In_map (f x) (fun y Hy => fin_tree (Acc_inv ax _ Hy))).
  Proof.
    unfold fin_tree at 1.
    rewrite Fix_F_equiv with (E := fun _ => eq); auto.
    intros; f_equal; apply list_In_map_ext.
    intros; auto.
  Qed.

  Fact fin_tree_root x Hx : tree_root (@fin_tree x Hx) = x.
  Proof.
    rewrite fin_tree_fix; simpl; auto.
  Qed.

  Fact fin_tree_sons x Hx : map (@tree_root _) (tree_sons (@fin_tree x Hx)) = f x. 
  Proof.
    rewrite fin_tree_fix; simpl.
    rewrite map_list_In_map.
    rewrite <- (map_id (f x)).
    rewrite <- list_In_map_eq_map.
    apply list_In_map_ext.
    intros; apply fin_tree_root.
  Qed.

  Definition fin_size : forall x, Acc (fun a b => In a (f b)) x -> nat. 
  Proof.
    apply Fix_F with (P := fun _ => nat).
    intros x Hx.
    exact (S (lsum (list_In_map _ Hx))).
  Defined.

  (* The height does not depend on the proof of accessibility *)

  Fact fin_size_eq x a1 a2 : @fin_size x a1 = @fin_size x a2.
  Proof.
    apply Fix_F_ext with (P := fun _ => nat) (E := fun _ => eq).
    intros; do 2 f_equal.
    apply list_In_map_ext; auto.
  Qed.

  Fact fin_size_fix x ax : 
     @fin_size x ax = S (lsum (list_In_map (f x) (fun y Hy => fin_size (Acc_inv ax _ Hy)))) . 
  Proof.
    unfold fin_size at 1.
    rewrite Fix_F_equiv with (E := fun _ => eq); auto.
    intros; do 2 f_equal; apply list_In_map_ext; auto.
  Qed.

  Fact fin_tree_size_eq x ax : tree_size (@fin_tree x ax) = fin_size ax.
  Proof.
    revert ax.
    induction ax as [ x ax IH ] using Acc_drect.
    rewrite fin_tree_fix, tree_size_fix, fin_size_fix.
    do 2 f_equal; rewrite map_list_In_map.
    apply list_In_map_ext.
    intros; auto.
  Qed.

End finitary_tree.

Section ft_rel.

  Variable (X : Type) (R : X -> X -> Prop) (f : X -> list X)
           (Hf : forall x y, R x y <-> In x (f y)).

  Let acc_eq x : Acc R x <-> Acc (fun x y => In x (f y)) x.
  Proof.
    split; apply Acc_incl; intros ? ?; apply Hf. 
  Qed.

  Definition ft_rel : forall x, Acc R x -> tree X.
  Proof.
    intros x Hx; apply fin_tree with (f := f) (x := x), acc_eq, Hx.
  Defined.

  Fact ft_rel_eq x H1 H2 : @ft_rel x H1 = @ft_rel x H2.
  Proof. apply fin_tree_eq. Qed.

  Fact ft_rel_root x Hx : tree_root (@ft_rel x Hx) = x.
  Proof. apply fin_tree_root. Qed.

  Fact ft_rel_sons x Hx : map (@tree_root _) (tree_sons (@ft_rel x Hx)) = f x. 
  Proof. apply fin_tree_sons. Qed.

End ft_rel. 

Section finitary_normal_form.

  Variable (X : Type) (R : X -> X -> Prop) (HR : forall x, { l | forall y, R x y <-> In y l }).
  
  Definition finitary_red_or_normal x : { y | R x y } + { forall y, ~ R x y }.
  Proof.
    destruct (HR x) as ([ | y l ] & H).
    right; intro; rewrite H; cbv; auto.
    left; exists y; apply H; left; auto.
  Qed.
  
  Definition finitary_normal_form x : Acc (fun x y => R y x) x -> { nx | clos_refl_trans _ R x nx /\ forall y, ~ R nx y }. 
  Proof.
    induction 1 as [ x Hx IHx ].
    destruct (finitary_red_or_normal x) as [ (y & Hy) | H ].
    destruct (IHx _ Hy) as (nx & H1 & H2).
    exists nx; split; auto.
    constructor 3 with y; auto; constructor 1; auto.
    exists x; split; auto.
    constructor 2.
  Qed.

End finitary_normal_form.

Section finitary_bound.

  Variable (X : Type) (R : X -> X -> Prop).

  Theorem bound_is_Acc x : (exists n, forall i y, rel_iter R i x y -> i < n) -> Acc (fun x y => R y x) x.
  Proof.
    intros (n & Hn).
    revert x Hn.
    induction n as [ | n IHn ]; intros x Hn.
    specialize (Hn 0 x eq_refl); omega.
    constructor; intros y Hy.
    apply IHn.
    intros i z Hz.
    apply lt_S_n, Hn with (y := z).
    exists y; split; auto.
  Qed.

End finitary_bound.

Require Import Arith.

Section finitary_compute.

  Variable (X Y : Type)
           (mX : X -> nat) (mY : Y -> nat) (HmY : forall n, { l | forall x, mY x < n -> In x l }).
           
  Let HmY' : forall n, { l | forall x, mY x < n <-> In x l }.
  Proof.
    intros n; destruct (HmY n) as (l & Hl).
    exists (filter (fun x => if le_lt_dec n (mY x) then false else true) l).
    intros x; split; rewrite filter_In.
    split; destruct (le_lt_dec n (mY x)); auto; omega.
    intros []; destruct (le_lt_dec n (mY x)); auto; discriminate 2.
  Qed.
  
  Variable   R : X -> Y -> Prop. 
  Hypothesis R_dec : forall x y, { R x y } + { ~ R x y }.
  Hypothesis R_bounded : forall n, { p | forall x y, R x y -> mX x < n -> mY y < p }.
  
  Definition finitary_compute x : (exists y, R x y) -> { y | R x y }.
  Proof.
    destruct (R_bounded (S (mX x))) as (p & Hp).
    destruct (HmY' p) as (lp & Hlp).
    set (lq := filter (fun y => if R_dec x y then true else false) lp).
    assert (forall y, R x y <-> In y lq) as Hlq.
      unfold lq.
      intros y; rewrite filter_In; destruct (R_dec x y).
      split; auto; split; auto.
      apply Hlp, Hp with x; auto.
      split; try tauto; intros []; discriminate 2.
    intros H1.
    assert (lq <> nil) as H2.
      destruct H1 as (y & Hy).
      apply Hlq in Hy.
      generalize lq Hy; intros [ [] | ]; discriminate.
    destruct lq as [ | y ? ].
    destruct H2; auto.
    exists y; apply Hlq; left; auto.
  Qed.
  
End finitary_compute.

    
    

     
