(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Omega Relations.

Require Import notations.

Set Implicit Arguments.

Fact inc2_refl X Y (R : rel X Y) : R inc2 R.
Proof. auto. Qed.

Fact inc2_trans X Y (R S T : rel X Y) : R inc2 S -> S inc2 T -> R inc2 T.
Proof. auto. Qed.

Definition total X Y (R : X -> Y -> Prop) : Prop := forall v, exists x, R v x.
Definition functional X Y (R : X -> Y -> Prop) : Prop := forall v x y, R v x -> R v y -> x = y.
Definition rel_equal X Y (R1 R2 : X -> Y -> Prop) := forall v x, R1 v x <-> R2 v x.
  
Infix "~req" := rel_equal (at level 70).  

Fact req_refl X Y (R : X -> Y -> Prop) : R ~req R.
Proof.
  red; tauto.
Qed.

Hint Resolve req_refl.
  
Fact req_sym X Y (R1 R2 : X -> Y -> Prop) : R1 ~req R2 -> R2 ~req R1.
Proof.
  unfold rel_equal; firstorder.
Qed.

Fact req_trans X Y (R1 R2 R3 : X -> Y -> Prop) : R1 ~req R2 -> R2 ~req R3 -> R1 ~req R3.
Proof.
  unfold rel_equal.
  intros H ? v x; rewrite H; auto.
Qed.
  
Fact req_functional X Y (R1 R2 : X -> Y -> Prop) : R1 ~req R2 -> functional R1 -> functional R2.
Proof.
  intros H H1 v ? ?; do 2 rewrite <- (H v _); apply H1.
Qed.
  
Fact req_total X Y (R1 R2 : X -> Y -> Prop) : R1 ~req R2 -> total R1 -> total R2.
Proof.
  intros H H1 v.
  destruct (H1 v) as (x & ?). 
  exists x; apply H; auto.
Qed.

Section closure_under_map.

  Variables (X Y : Type) (f : X -> Y).  
  Variable R : rel X X.
  Variable S : rel Y Y.
  
  Hypothesis H : forall x y, R x y -> S (f x) (f y).

  Fact clos_trans_map x y : clos_trans _ R x y -> clos_trans _ S (f x) (f y).
  Proof.
    induction 1.
    constructor 1; apply H; auto.
    constructor 2 with (f y); auto.
  Qed.
  
  Fact clos_refl_trans_map x y : clos_refl_trans _ R x y -> clos_refl_trans _ S (f x) (f y).
  Proof.
    induction 1.
    constructor 1; apply H; auto.
    constructor 2.
    constructor 3 with (f y); auto.
  Qed.

  Hypothesis Hrev : forall x z, S (f x) z -> exists y, z = f y /\ clos_refl_trans _ R x y.

  Let clos_refl_trans_map_rev_r u v : clos_refl_trans _ S u v 
              -> forall x, u = f x -> exists y, v = f y /\ clos_refl_trans _ R x y.
  Proof.
    induction 1 as [ u v Huv | u | u v w H1 IH1 H2 IH2 ]; intros x H3; subst.
    apply Hrev in Huv; destruct Huv as (y & ? & Huv); subst.
    exists y; split; auto; constructor 1; auto.
    exists x; split; auto; constructor 2; auto.
    destruct (IH1 _ eq_refl) as (y & ? & Hy); subst.
    destruct (IH2 _ eq_refl) as (z & ? & Hz); subst.
    exists z; split; auto; constructor 3 with y; auto.
  Qed.
  
  Fact clos_refl_trans_map_rev x v : clos_refl_trans _ S (f x) v -> exists y, v = f y /\ clos_refl_trans _ R x y.
  Proof.
    intros H1.
    apply clos_refl_trans_map_rev_r with (1 := H1); auto.
  Qed.

  Fact clos_refl_sym_trans_map x y : clos_refl_sym_trans _ R x y -> clos_refl_sym_trans _ S (f x) (f y).
  Proof.
    induction 1.
    constructor 1; apply H; auto.
    constructor 2.
    constructor 3; auto.
    constructor 4 with (f y); auto.
  Qed.
  
  Fact clos_refl_trans_step x y : clos_refl_trans _ R x y -> x = y \/ exists z, R x z /\ clos_refl_trans _ R z y.
  Proof.
    intros H1.
    apply clos_rt_rt1n in H1.
    destruct H1 as [ | y ]; [ left | right ]; auto.
    exists y; split; auto; apply clos_rt1n_rt; auto.
  Qed.

  Fact clos_refl_trans_step' x y : clos_refl_trans _ R y x -> x = y \/ exists z, R z x /\ clos_refl_trans _ R y z.
  Proof.
    intros H1.
    apply clos_rt_rtn1 in H1.
    destruct H1 as [ | x ]; [ left | right ]; auto.
    exists x; split; auto; apply clos_rtn1_rt; auto.
  Qed.

  Hypothesis Rfun : forall x y z, R x y -> R x z -> y = z. 

  Fact clos_refl_trans_fun_cr x y z : 
      clos_refl_trans _ R x y 
   -> clos_refl_trans _ R x z 
   -> clos_refl_trans _ R y z
   \/ clos_refl_trans _ R z y.
  Proof.
    induction 1 as [ y x H0 | x | u y x H1 IH1 H2 IH2 ]; intros H3.
    apply clos_refl_trans_step in H3; destruct H3 as [ H3 | (v & H3 & ?) ].
    right; subst; constructor 1; auto.
    generalize (Rfun H0 H3); intros; subst; left; auto.
    left; auto.
    apply IH1 in H3; destruct H3 as [ H3 | H3 ].
    destruct (IH2 H3) as [ H4 | H4 ].
    left; auto.
    right; auto.
    right.
    constructor 3 with (1 := H3); auto.
  Qed.

End closure_under_map.

Section closure_mono.

  Variables (X : Type) (R S : rel X X) (H : R inc2 S).

  Fact clos_trans_mono : clos_trans _ R inc2 clos_trans _ S.
  Proof.
    induction 1.
    constructor 1; auto.
    constructor 2 with y; auto.
  Qed.
  
  Fact clos_refl_trans_mono : clos_refl_trans _ R inc2 clos_refl_trans _ S.
  Proof.
    induction 1.
    constructor 1; apply H; auto.
    constructor 2.
    constructor 3 with y; auto.
  Qed.

  Fact clos_refl_sym_trans_mono : clos_refl_sym_trans _ R inc2 clos_refl_sym_trans _ S.
  Proof.
    induction 1.
    constructor 1; apply H; auto.
    constructor 2.
    constructor 3; auto.
    constructor 4 with y; auto.
  Qed.

End closure_mono.

Section closure_inc.

  Variables (X : Type) (R : rel X X).

  Fact clos_trans_inc : R inc2 clos_trans _ R.
  Proof.
    constructor 1; auto.
  Qed.

  Fact clos_refl_trans_inc : R inc2 clos_refl_trans _ R.
  Proof.
    constructor 1; auto.
  Qed.
  
  Fact clos_refl_sym_trans_inc : R inc2 clos_refl_sym_trans _ R.
  Proof.
    constructor 1; auto.
  Qed.

End closure_inc.

Section closure_idem.
  
  Variables (X : Type) (R : rel X X).

  Fact clos_trans_idem : clos_trans _ (clos_trans _ R) inc2 clos_trans _ R.
  Proof.
    induction 1; auto.
    constructor 2 with y; auto.
  Qed.

  Fact clos_refl_trans_idem : clos_refl_trans _ (clos_refl_trans _ R) inc2 clos_refl_trans _ R.
  Proof.
    induction 1; auto.
    constructor 2.
    constructor 3 with y; auto.
  Qed.
  
  Fact clos_refl_sym_trans_idem : clos_refl_sym_trans _ (clos_refl_sym_trans _ R) inc2 clos_refl_sym_trans _ R.
  Proof.
    induction 1; auto.
    constructor 2.
    constructor 3; auto.
    constructor 4 with y; auto.
  Qed.

End closure_idem.

Section clos_refl.

  Variables (X : Type) (R : rel X X).
  
  Definition clos_refl := eq cup2 R.
  
  Fact clos_refl_rt x y : clos_refl_trans _ R x y <-> clos_refl_trans _ clos_refl x y.
  Proof.
    split.
    apply clos_refl_trans_mono.
    intros; right; auto.
    induction 1.
    destruct H.
    subst; constructor 2.
    constructor 1; auto.
    constructor 2.
    constructor 3 with y; auto.
  Qed.

End clos_refl.

Section clos_trans_refl_trans.

  Variables (X : Type) (R : rel X X).

  Fact clos_t_incl_rt : clos_trans _ R inc2 clos_refl_trans _ R.
  Proof.
    induction 1.
    constructor 1; auto.
    constructor 3 with y; auto.
  Qed.
  
  Fact clos_rt_t_eq x y : clos_trans _ R x y <-> exists z, R z y /\ clos_refl_trans _ R x z.
  Proof.
    split.
    induction 1.
    exists x; split; auto; constructor 2.
    destruct IHclos_trans2 as (h & ? & ?).
    exists h; split; auto.
    constructor 3 with y; auto.
    apply clos_t_incl_rt; auto.
    intros (z & H1 & H2).
    apply clos_rt_rtn1_iff in H2.
    revert y H1.
    induction H2; intros.
    constructor 1; auto.
    constructor 2 with z.
    apply IHclos_refl_trans_n1; auto.
    constructor 1; auto.
  Qed.

  Fact clos_rt_t_eq' x y : clos_trans _ R x y <-> exists z, R x z /\ clos_refl_trans _ R z y.
  Proof.
    split.
    induction 1.
    exists y; split; auto; constructor 2.
    destruct IHclos_trans1 as (h & ? & ?).
    exists h; split; auto.
    constructor 3 with y; auto.
    apply clos_t_incl_rt; auto.
    intros (z & H1 & H2).
    apply clos_rt_rt1n_iff in H2.
    revert x H1.
    induction H2; intros.
    constructor 1; auto.
    constructor 2 with x.
    constructor 1; auto.
    apply IHclos_refl_trans_1n; auto.
  Qed.

End clos_trans_refl_trans.
