(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Max Omega.

Require Import utils.

Set Implicit Arguments.

Section trees.

  Variable (X : Type).

  (* we do not want the too weak Coq generated induction principles *)

  Unset Elimination Schemes.

  Inductive tree : Type := in_tree : X -> list tree -> tree.

  Set Elimination Schemes.

  Definition tree_root t := match t with in_tree x _ => x end.
  Definition tree_sons t := match t with in_tree _ l => l end.

  Fact tree_root_sons_eq t : t = in_tree (tree_root t) (tree_sons t).
  Proof. destruct t; auto. Qed.
  
  (* the immediate subtree relation *)
  
  Definition imsub_tree s t := match t with in_tree _ ll => In s ll end.
  
  Infix "<ist" := imsub_tree (at level 70).
  
  Fact imsub_tree_fix s t :  s <ist t <-> exists x ll, In s ll /\ t = in_tree x ll.
  Proof.
    split.
    destruct t as [ x ll ]; exists x, ll; auto.
    intros (x & ll & H1 & ?); subst; auto.
  Qed.
  
  (* The immediate subtree relation is well founded *)
  
  Fixpoint imsub_tree_wf t : Acc imsub_tree t.
  Proof.
    refine (
      match t with
        | in_tree x ll => Acc_intro _ _
      end); simpl; clear x t.
    induction ll as [ | x ll IH ].
    intros _ [].
    intros ? [ [] | ].
    apply imsub_tree_wf.
    apply IH.
  Qed.

  (* let us define our own induction principles *)

  Section tree_rect.

    Variable P : tree -> Type.
    Hypothesis f : forall a ll, (forall x, In x ll -> P x) -> P (in_tree a ll).

    Let g : forall t, (forall s, imsub_tree s t -> P s) -> P t.
    Proof.  intros [] H; apply f, H. Defined.

    Definition tree_rect t : P t.
    Proof.
      apply Fix with (1 := imsub_tree_wf), g.
    Defined.

    Variable (E : forall t, P t -> P t -> Prop).

    Hypothesis f_ext : forall a ll f1 f2, (forall x Hx, E (f1 x Hx) (f2 x Hx)) 
                                        -> E (f a ll f1) (f a ll f2).

    Let Hg : forall t f1 f2, (forall s Hs, E (f1 s Hs) (f2 s Hs)) -> E (g t f1) (g t f2).
    Proof.
      intros [ ]; apply f_ext.
    Qed. 

    (* Coq recursive type-checking does not allow such a definition 
       but it is possible to prove this identity
    *)

    Fact tree_rect_fix a ll : E (tree_rect (in_tree a ll)) (f a ll (fun t _ => tree_rect t)).
    Proof.
      apply Fix_equiv with (1 := Hg).
    Qed.

  End tree_rect.
  
  Definition tree_rec (P : tree -> Set)  := tree_rect P.
  Definition tree_ind (P : tree -> Prop) := tree_rect P.

  Section tree_recursion.

    (* the particular case when the output type does not depend on the tree *)

    Variables (Y : Type) (f : X -> list tree -> list Y -> Y).    
  
    Definition tree_recursion : tree -> Y.
    Proof.
      apply tree_rect.
      intros x ll IH.
      apply (f x ll (list_In_map _ IH)).
    Defined.
    
    (* In that case, extensionnality is for free *)

    Fact tree_recursion_fix x ll : tree_recursion (in_tree x ll) = f x ll (map tree_recursion ll).
    Proof.
      unfold tree_recursion at 1.
      rewrite tree_rect_fix with (E := fun _ => eq).
      f_equal.
      induction ll; simpl; f_equal; auto.
      clear x ll; intros x ll g h H; simpl.
      f_equal; apply list_In_map_ext, H.
    Qed.
  
  End tree_recursion.

  Definition tree_ht : tree -> nat.
  Proof.
    induction 1 as [ _ _ IH ] using tree_recursion.
    apply (S (lmax IH)).
  Defined.

  Definition tree_size : tree -> nat.
  Proof.
    induction 1 as [ _ _ IH ] using tree_recursion.
    apply (S (lsum IH)).
  Defined.
  
  Fact tree_ht_fix x ll : tree_ht (in_tree x ll) = S (lmax (map tree_ht ll)).
  Proof.
    apply tree_recursion_fix.
  Qed.

  Fact tree_size_fix x ll : tree_size (in_tree x ll) = S (lsum (map tree_size ll)).
  Proof.
    apply tree_recursion_fix.
  Qed.
  
  Fact tree_ht_0 t : 0 < tree_ht t.
  Proof.
    destruct t; rewrite tree_ht_fix; omega.
  Qed.
  
  Fact tree_ht_find x ll : ll <> nil -> { t | In t ll /\ tree_ht (in_tree x ll) = S (tree_ht t) }.
  Proof.
    intros H.
    destruct (lmax_inv (map tree_ht ll)) as [ H1 | H1 ].
    exfalso; destruct ll.
    apply H; auto.
    discriminate H1.
    apply In_map_nat_find in H1.
    destruct H1 as (t & H1 & H2).
    exists t; split; auto.
    rewrite tree_ht_fix, H1; auto.
  Qed.

End trees.