(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Omega.

Require Import utils pos vec.
Require Import rec_alg.

Require Import uterm uterm_beta uterm_beta_eq.
Require Import uterm_normal uterm_head uterm_head_sem uterm_nat uterm_repr.

Set Implicit Arguments.

(** Every recursive algorithm f of arity k that defines 
    a "provably total" (functional) relation Rf = recalg_rel f
    
      Rf : vec nat k -> nat -> Prop
 
   can be transformed into a (Coq definable) 
   function 
 
      cf : vec nat k -> nat 

   which reifies (or computes the values of) the
   relation Rf, ie Rf v (cf v) holds for any v

   "Provably total" can be understood either without axioms 
   or with axioms, it does not matter.

   Of course to compute cf, we need a representation of Rf
   such as the algorithm f. If, for a given Rf, we only a
   have a proof that an algorithm f for Rf exists, then 
   we can only prove that cf exists, but we lack info to
   compute it in that case. 
   
   
  Orelse, total recursive functions can be reified in Coq
**)

Local Notation " '[|' f '|]' " := (@recalg_rel _ f) (at level 0).

Theorem Coq_is_total k (f : recalg k) : (forall v, exists x, [|f|] v x) -> { cf | forall v, [|f|] v (cf v) }.
Proof.
  intros H.
  destruct (ut_repr_recalg 0 f) as (tf & Htf).
  set (t v := tf @* map (ut_nat 0) (@vec_list _ k v)).
  assert (forall v, ut_normalizable (t v)) as Ht.
    intros v; unfold t.
    destruct (H v) as (x & Hx).
    apply Htf in Hx.
    generalize (@ut_nat_normal 0 x); intros H1.
    exists (ut_nat 0 x); split; auto.
    revert H1 Hx.
    apply ut_beta_normal_eq.
  set (H2 v := ut_normalizable_nf (Ht v)).
  set (nf v := proj1_sig (H2 v)).
  assert (Hnf : forall v, t v -l>> nf v /\ ut_normal (nf v)).
    intros v; apply (proj2_sig (H2 v)).
  generalize nf Hnf; clear Ht nf Hnf H2; intros nf Hnf.
  set (cf v := ut_size (nf v) - 3).
  assert (forall v, nf v = ut_nat _ (cf v)) as Hcf.
    intros v.
    destruct (H v) as (x & Hx).
    apply Htf in Hx.
    destruct (Hnf v) as [ H2 H3 ].
    unfold t in H2.
    apply ut_leftmost_rt_beta_rt, ut_beta_rt_eq, ut_beta_eq_sym in H2.
    generalize (ut_beta_eq_trans H2 Hx).
    intros H4.
    apply ut_beta_eq_normal in H4; auto.
    unfold cf.
    rewrite H4, ut_nat_size; f_equal; omega.

  exists cf.
  intros v.
  apply Htf, ut_beta_rt_eq, ut_leftmost_rt_beta_rt.
  rewrite <- Hcf; apply Hnf.
Qed.

Check Coq_is_total.
Print Assumptions Coq_is_total.

Section nat_pair.
  
  Let a_fst := proj1_sig rf_prim_fst.
  Let a_snd := proj1_sig rf_prim_snd.
  Let a_pair := proj1_sig rf_prim_pair.
  
  Definition ut_nat_fst_0  := proj1_sig (ut_repr_recalg 0 a_fst).
  Definition ut_nat_snd_0  := proj1_sig (ut_repr_recalg 0 a_snd).
  Definition ut_nat_pair_0 := proj1_sig (ut_repr_recalg 0 a_pair).
  
  Let nat_fst_prop v x : x = nat_fst v <-> ut_nat_fst_0 @ ut_nat _ v ~b ut_nat _ x.
  Proof.
    rewrite <- (proj2_sig (ut_repr_recalg _ a_fst) (v##vec_nil) x).
    unfold a_fst; rewrite <- (proj2 (proj2_sig rf_prim_fst) (v##vec_nil) x).
    rewrite (s_repr_fst (v##vec_nil) x).
    simpl; tauto.
  Qed.
  
  Let nat_snd_prop v x : x = nat_snd v <-> ut_nat_snd_0 @ ut_nat _ v ~b ut_nat _ x.
  Proof.
    rewrite <- (proj2_sig (ut_repr_recalg _ a_snd) (v##vec_nil) x).
    unfold a_snd; rewrite <- (proj2 (proj2_sig rf_prim_snd) (v##vec_nil) x).
    rewrite (s_repr_snd (v##vec_nil) x).
    simpl; tauto.
  Qed.
  
  Let nat_pair_prop v w x : x = nat_pair v w <-> ut_nat_pair_0 @ ut_nat _ v @ ut_nat _ w ~b ut_nat _ x.
  Proof.
    rewrite <- (proj2_sig (ut_repr_recalg _ a_pair) (v##w##vec_nil) x).
    unfold a_pair; rewrite <- (proj2 (proj2_sig rf_prim_pair) (v##w##vec_nil) x).
    rewrite (s_repr_pair (v##w##vec_nil) x).
    simpl; tauto.
  Qed.
  
  Fact ut_nat_fst_0_prop v : ut_nat_fst_0 @ ut_nat _ v -b>> ut_nat _ (nat_fst v).
  Proof.
    apply ut_beta_normal_eq.
    apply ut_nat_normal.
    apply nat_fst_prop; auto.
  Qed.
  
  Fact ut_nat_snd_0_prop v : ut_nat_snd_0 @ ut_nat _ v -b>> ut_nat _ (nat_snd v).
  Proof.
    apply ut_beta_normal_eq.
    apply ut_nat_normal.
    apply nat_snd_prop; auto.
  Qed.
  
  Fact ut_nat_pair_0_prop v w : ut_nat_pair_0 @ ut_nat _ v @ ut_nat _ w -b>> ut_nat _ (nat_pair v w).
  Proof.
    apply ut_beta_normal_eq.
    apply ut_nat_normal.
    apply nat_pair_prop; auto.
  Qed.
  
  Definition ut_lift_rec (t : uterm 0) n : uterm n := ut_subst t (fun _ => \\ £ pos0).
  
  Fact ut_lift_rec_lift n t : ut_lift (ut_lift_rec t n) = ut_lift_rec t (S n).
  Proof.
    unfold ut_lift_rec.
    rewrite ut_lift_subst_1.
    apply ut_subst_ext.
    intros p; pos_inv p.
  Qed.
  
  Fact ut_lift_rec_subst n t m (s : uts_t n m) : ut_subst (ut_lift_rec t n) s = ut_lift_rec t m.
  Proof.
    unfold ut_lift_rec.
    rewrite ut_subst_comp.
    apply ut_subst_ext.
    intros p; pos_inv p.
  Qed.
  
  Variable n : nat.

  Definition ut_nat_fst  := ut_lift_rec ut_nat_fst_0 n.
  Definition ut_nat_snd  := ut_lift_rec ut_nat_snd_0 n.
  Definition ut_nat_pair := ut_lift_rec ut_nat_pair_0 n.
  
  Fact ut_lift_rec_nat k : ut_nat n k = ut_lift_rec (ut_nat 0 k) n.
  Proof.
    unfold ut_lift_rec; rewrite ut_nat_subst; auto.
  Qed.
  
  Fact ut_lift_rec_app x y : ut_lift_rec x n @ ut_lift_rec y n = ut_lift_rec (x @ y) n.
  Proof. auto. Qed.
 
  Fact ut_nat_fst_prop v : ut_nat_fst @ ut_nat _ v -b>> ut_nat _ (nat_fst v).
  Proof.
    unfold ut_nat_fst.
    do 2 rewrite ut_lift_rec_nat.
    rewrite ut_lift_rec_app.
    apply ut_beta_rt_subst, ut_nat_fst_0_prop.
  Qed.
  
  Fact ut_nat_snd_prop v : ut_nat_snd @ ut_nat _ v -b>> ut_nat _ (nat_snd v).
  Proof.
    unfold ut_nat_snd.
    do 2 rewrite ut_lift_rec_nat.
    rewrite ut_lift_rec_app.
    apply ut_beta_rt_subst, ut_nat_snd_0_prop.
  Qed.
  
  Fact ut_nat_pair_prop v w : ut_nat_pair @ ut_nat _ v @ ut_nat _ w -b>> ut_nat _ (nat_pair v w).
  Proof.
    unfold ut_nat_pair.
    do 3 rewrite ut_lift_rec_nat.
    do 2 rewrite ut_lift_rec_app.
    apply ut_beta_rt_subst, ut_nat_pair_0_prop.
  Qed.

End nat_pair.

Fact ut_nat_fst_subst n m s : @ut_subst n (ut_nat_fst _) m s = ut_nat_fst m.
Proof. apply ut_lift_rec_subst. Qed.

Fact ut_nat_snd_subst n m s : @ut_subst n (ut_nat_snd _) m s = ut_nat_snd m.
Proof. apply ut_lift_rec_subst. Qed.

Fact ut_nat_pair_subst n m s : @ut_subst n (ut_nat_pair _) m s = ut_nat_pair m.
Proof. apply ut_lift_rec_subst. Qed.

Fact ut_nat_fst_lift n : ut_lift (ut_nat_fst n) = ut_nat_fst _.
Proof. apply ut_lift_rec_lift. Qed.

Fact ut_nat_snd_lift n : ut_lift (ut_nat_snd n) = ut_nat_snd _.
Proof. apply ut_lift_rec_lift. Qed.

Fact ut_nat_pair_lift n : ut_lift (ut_nat_pair n) = ut_nat_pair _.
Proof. apply ut_lift_rec_lift. Qed.
