(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Omega.

Require Import utils pos vec.

Require Import uterm uterm_beta uterm_beta_eq uterm_beta_confluent.
Require Import uterm_normal uterm_head uterm_solvable uterm_head_sem.
Require Import uterm_bool.

Set Implicit Arguments.

Local Hint Resolve ut_true_normal ut_false_normal.

Local Fact ut_not_true_and_false k f : f -b>> ut_true k -> f -b>> ut_false _ -> False.
Proof.
  intros H1 H2.
  apply ut_beta_rt_eq in H2.
  apply ut_beta_rt_eq, ut_beta_eq_sym in H1.
  generalize (ut_beta_eq_trans H1 H2); intros H3.
  apply ut_beta_eq_normal in H3; auto.
  discriminate H3.
Qed.

Definition ut_separator k (P Q : uterm k -> Prop) (s : uterm k) :=
    forall t, ( s @ t -b>> ut_true _ \/ s @ t -b>> ut_false _ )
           /\ ( P t -> s @ t -b>> ut_true _ )
           /\ ( Q t -> s @ t -b>> ut_false _ ).

(* The is something close to Theorem 2.20 p 38 of Krivine-Cori *)

Section uterm_undec.

  (* uterm k is partitionned in a non trivial way into S + ~ S 
     with P inc S and Q inc ~ S and both P and Q are ~eq stable
     and non empty  
  *)

  Variable (k : nat) (P Q : uterm k -> Prop) (HPQ : sig (ut_separator P Q)).

  Hypothesis H2 : exists p, P p.
  Hypothesis H3 : exists q, Q q.
  Hypothesis H4 : forall t1 t2, t1 ~b t2 -> P t1 -> P t2.
  Hypothesis H5 : forall t1 t2, t1 ~b t2 -> Q t1 -> Q t2.
  
  Theorem ut_beta_eq_stable_and_decidable_partitions_are_trivial : False.
  Proof.
    destruct H2 as (p & Hp).
    destruct H3 as (q & Hq).
    clear H2 H3.
    revert HPQ; intros (s & Hs).
    red in Hs.
    
    (* We use the trick of applying a term to itself provoking a contradiction
       which in this case, is implemented by a swap between p nd q *)
    
    set (t := \\ (ut_lift s @ (£ pos0 @ £ pos0) @ ut_lift q @ ut_lift p)).
    destruct (proj1 (Hs (t @ t))) as [ H | H ].
    
    (**     s @ (t @ t) -b>> ut_true
       --> t @ t -b>> q 
       --> t @ t ~b q 
       --> Q (t @ t)
       --> s @ (t @ t) -b>> ut_false 
     **)
     
    apply ut_not_true_and_false with (1 := H).
    apply Hs, H5 with (2 := Hq).
    apply ut_beta_eq_sym, ut_beta_rt_eq.
    unfold t at 1; ut_beta_rt_step.
    repeat rewrite ut_reduce_app.
    repeat rewrite ut_reduce_lift.
    repeat rewrite ut_reduce_pos0.
    apply ut_beta_rt_trans with (ut_true _ @ q @ p).
    do 2 apply ut_beta_rt_lft; auto.
    apply ut_whead_rt_beta_rt, ut_true_prop.
    
    (**     s @ (t @ t) -b>> ut_false
       --> t @ t -b>> p 
       --> t @ t ~b p 
       --> P (t @ t)
       --> s @ (t @ t) -b>> ut_true
     **)
     
    apply ut_not_true_and_false with (2 := H).
    apply Hs, H4 with (2 := Hp).
    apply ut_beta_eq_sym, ut_beta_rt_eq.
    unfold t at 1; ut_beta_rt_step.
    repeat rewrite ut_reduce_app.
    repeat rewrite ut_reduce_lift.
    repeat rewrite ut_reduce_pos0.
    apply ut_beta_rt_trans with (ut_false _ @ q @ p).
    do 2 apply ut_beta_rt_lft; auto.
    apply ut_whead_rt_beta_rt, ut_false_prop.
  Qed.

End uterm_undec.

(* A predicate P over uterm k is decidable if there exists a term d which decides P *)

Definition ut_decidable k (P : uterm k -> Prop) :=
   exists d, forall t, (d @ t -b>> ut_true _ \/ d @ t -b>> ut_false _) /\ (P t <-> d @ t -b>> ut_true _).
                                                              
Fact ut_decidable_decider k P : @ut_decidable k P <-> (forall t, P t \/ ~ P t) /\ exists s, ut_separator P (fun t => ~ P t) s.
Proof.
  split.
  
  intros (d & Hd); split.
  
  intros t.
  rewrite (proj2 (Hd t)).
  destruct (proj1 (Hd t)); auto.
  right; intro; apply ut_not_true_and_false with (2 := H); auto.
  
  exists d; intros t; split.
  
  apply Hd.
  split.
  rewrite (proj2 (Hd t)); auto.
  rewrite (proj2 (Hd t)).
  intros H.
  destruct (proj1 (Hd t)); tauto.
  
  intros (H & s & Hs); exists s; intros t; split; red in Hs.
  
  apply Hs.
  split.
  apply Hs.
  intros H1.
  destruct (H t) as [ | C ]; auto.
  apply Hs in C.
  exfalso; revert H1 C; apply ut_not_true_and_false.
Qed.

Fact ut_id_solvable k : ut_solvable (ut_id k).
Proof.
  apply ut_abs_solvable, ut_var_solvable.
Qed.

Fact ut_id_normalizable k : ut_normalizable (ut_id k).
Proof.
  apply ut_normal_normalizable, ut_abs_normal, ut_var_normal.
Qed.

Definition ut_delta k : uterm k := \\ (£ pos0 @ £ pos0).

Fact ut_delta_delta_not_solvable k : ~ ut_solvable (ut_delta k @ ut_delta _).
Proof.
  rewrite ut_solvable_head_sn.
  apply Acc_irrefl.
  do 2 constructor 1.
Qed.

Fact ut_delta_delta_not_normalizable k : ~ ut_normalizable (ut_delta k @ ut_delta _).
Proof.
  intros H; apply ut_normalizable_solvable in H; revert H.
  apply ut_delta_delta_not_solvable.
Qed.

Hint Resolve ut_id_solvable ut_id_normalizable ut_delta_delta_not_solvable ut_delta_delta_not_normalizable.

Fact ut_beta_eq_normalizable k (u v : uterm k) : u ~b v -> ut_normalizable u -> ut_normalizable v.
Proof.
  intros H1 (t & H2 & H3); exists t; split; auto.
  apply ut_beta_normal_eq; auto.
  apply ut_beta_eq_trans with u.
  apply ut_beta_eq_sym; auto.
  apply ut_beta_rt_eq; auto.
Qed.

(* The set of normalizable terms cannot be decided by a lambda term *)

Corollary ut_normalizable_undec k : ~ ut_decidable (@ut_normalizable k).
Proof.
  rewrite ut_decidable_decider.
  intros (H0 & s & Hs).
  apply ut_beta_eq_stable_and_decidable_partitions_are_trivial with (1 := exist _ s Hs).
  
  exists (@ut_id _); auto.
  exists (@ut_delta _ @ @ut_delta _); auto.
  apply ut_beta_eq_normalizable.
  intros ? ? ? H; contradict H; revert H.
  apply ut_beta_eq_normalizable, ut_beta_eq_sym; auto.
Qed.

(* The set of solvable terms cannot be decided by a lambda term *)

Corollary ut_solvable_undec k : ~ ut_decidable (@ut_solvable k).
Proof.
  rewrite ut_decidable_decider.
  intros (H0 & s & Hs).
  apply ut_beta_eq_stable_and_decidable_partitions_are_trivial with (1 := exist _ s Hs).
  
  exists (@ut_id _); auto.
  exists (@ut_delta _ @ @ut_delta _); auto.
  apply ut_beta_eq_solvable.
  intros ? ? ? H; contradict H; revert H.
  apply ut_beta_eq_solvable; auto.
Qed.

Check ut_normalizable_undec.
Print Assumptions ut_normalizable_undec.

Corollary ut_beta_eq_undec k f : ~ ut_decidable (@ut_beta_eq k f).
Proof.
  (* How do I build a term which is not ~b to f ? 
     If f does not contain pos0 (like ut_lift _) then £ pos0 is not ~b to f
     but what to do with closed terms or terms that use all their variables ? 
     
     I am not sure we can actually show that there exists a closed term which is 
     non beta equivalent to another closed term f intuitionistically. We can
     show that there are infinitelly many of them, but how to constructivelly
     build one which is not equivalent ?
     
     Is it possible to decide terms that do contain an occurence of pos0 ?
     This is not closed under beta equivalence so ...
  *)
Admitted.
